@@ -1,22 +1,36 @@
 import asyncio
 import functools
-import ssl
 import sys
 import traceback
 import warnings
 from collections import defaultdict
 from hashlib import md5, sha1, sha256
+from itertools import cycle, islice
+from time import monotonic
 from types import MappingProxyType
 
 from . import hdrs, helpers
-from .client_exceptions import (ClientConnectorError, ClientHttpProxyError,
+from .client_exceptions import (ClientConnectionError,
+                                ClientConnectorCertificateError,
+                                ClientConnectorError, ClientConnectorSSLError,
+                                ClientHttpProxyError,
                                 ClientProxyConnectionError,
-                                ServerFingerprintMismatch)
+                                ServerFingerprintMismatch, certificate_errors,
+                                ssl_errors)
 from .client_proto import ResponseHandler
 from .client_reqrep import ClientRequest
 from .helpers import SimpleCookie, is_ip_address, noop, sentinel
+from .locks import EventResultOrError
+from .log import client_logger
 from .resolver import DefaultResolver
 
+
+try:
+    import ssl
+except ImportError:  # pragma: no cover
+    ssl = None
+
+
 __all__ = ('BaseConnector', 'TCPConnector', 'UnixConnector')
 
 HASHFUNC_BY_DIGESTLEN = {
@@ -297,11 +311,11 @@ class BaseConnector(object):
             if self._loop.is_closed():
                 return noop()
 
-            # cacnel cleanup task
+            # cancel cleanup task
             if self._cleanup_handle:
                 self._cleanup_handle.cancel()
 
-            # cacnel cleanup close task
+            # cancel cleanup close task
             if self._cleanup_closed_handle:
                 self._cleanup_closed_handle.cancel()
 
@@ -335,7 +349,7 @@ class BaseConnector(object):
     @asyncio.coroutine
     def connect(self, req):
         """Get from pool or create new connection."""
-        key = (req.host, req.port, req.ssl)
+        key = req.connection_key
 
         if self._limit:
             # total calc available connections
@@ -361,10 +375,13 @@ class BaseConnector(object):
             # This connection will now count towards the limit.
             waiters = self._waiters[key]
             waiters.append(fut)
-            yield from fut
-            waiters.remove(fut)
-            if not waiters:
-                del self._waiters[key]
+            try:
+                yield from fut
+            finally:
+                # remove a waiter even if it was cancelled
+                waiters.remove(fut)
+                if not waiters:
+                    del self._waiters[key]
 
         proto = self._get(key)
         if proto is None:
@@ -373,14 +390,20 @@ class BaseConnector(object):
             self._acquired_per_host[key].add(placeholder)
             try:
                 proto = yield from self._create_connection(req)
-            except OSError as exc:
-                raise ClientConnectorError(
-                    exc.errno,
-                    'Cannot connect to host {0[0]}:{0[1]} ssl:{0[2]} [{1}]'
-                    .format(key, exc.strerror)) from exc
+                if self._closed:
+                    proto.close()
+                    raise ClientConnectionError("Connector is closed.")
+            except:
+                # signal to waiter
+                for waiter in self._waiters[key]:
+                    if not waiter.done():
+                        waiter.set_result(None)
+                        break
+                raise
             finally:
-                self._acquired.remove(placeholder)
-                self._acquired_per_host[key].remove(placeholder)
+                if not self._closed:
+                    self._acquired.remove(placeholder)
+                    self._acquired_per_host[key].remove(placeholder)
 
         self._acquired.add(proto)
         self._acquired_per_host[key].add(proto)
@@ -482,6 +505,55 @@ class BaseConnector(object):
 _SSL_OP_NO_COMPRESSION = getattr(ssl, "OP_NO_COMPRESSION", 0)
 
 
+class _DNSCacheTable:
+
+    def __init__(self, ttl=None):
+        self._addrs = {}
+        self._addrs_rr = {}
+        self._timestamps = {}
+        self._ttl = ttl
+
+    def __contains__(self, host):
+        return host in self._addrs
+
+    @property
+    def addrs(self):
+        return self._addrs
+
+    def add(self, host, addrs):
+        self._addrs[host] = addrs
+        self._addrs_rr[host] = cycle(addrs)
+
+        if self._ttl:
+            self._timestamps[host] = monotonic()
+
+    def remove(self, host):
+        self._addrs.pop(host, None)
+        self._addrs_rr.pop(host, None)
+
+        if self._ttl:
+            self._timestamps.pop(host, None)
+
+    def clear(self):
+        self._addrs.clear()
+        self._addrs_rr.clear()
+        self._timestamps.clear()
+
+    def next_addrs(self, host):
+        # Return an iterator that will get at maximum as many addrs
+        # there are for the specific host starting from the last
+        # not itereated addr.
+        return islice(self._addrs_rr[host], len(self._addrs[host]))
+
+    def expired(self, host):
+        if self._ttl is None:
+            return False
+
+        return (
+            self._timestamps[host] + self._ttl
+        ) < monotonic()
+
+
 class TCPConnector(BaseConnector):
     """TCP connector.
 
@@ -495,6 +567,7 @@ class TCPConnector(BaseConnector):
     resolver - Enable DNS lookups and use this
         resolver
     use_dns_cache - Use memory cache for DNS lookups.
+    ttl_dns_cache - Max seconds having cached a DNS entry, None forever.
     family - socket address family
     local_addr - local tuple of (host, port) to bind socket to
 
@@ -507,7 +580,7 @@ class TCPConnector(BaseConnector):
     """
 
     def __init__(self, *, verify_ssl=True, fingerprint=None,
-                 resolve=sentinel, use_dns_cache=True,
+                 resolve=sentinel, use_dns_cache=True, ttl_dns_cache=10,
                  family=0, ssl_context=None, local_addr=None,
                  resolver=None, keepalive_timeout=sentinel,
                  force_close=False, limit=100, limit_per_host=0,
@@ -531,10 +604,11 @@ class TCPConnector(BaseConnector):
             if not hashfunc:
                 raise ValueError('fingerprint has invalid length')
             elif hashfunc is md5 or hashfunc is sha1:
-                warnings.simplefilter('always')
                 warnings.warn('md5 and sha1 are insecure and deprecated. '
                               'Use sha256.',
                               DeprecationWarning, stacklevel=2)
+                client_logger.warn('md5 and sha1 are insecure and deprecated. '
+                                   'Use sha256.')
             self._hashfunc = hashfunc
         self._fingerprint = fingerprint
 
@@ -543,11 +617,19 @@ class TCPConnector(BaseConnector):
         self._resolver = resolver
 
         self._use_dns_cache = use_dns_cache
-        self._cached_hosts = {}
+        self._cached_hosts = _DNSCacheTable(ttl=ttl_dns_cache)
+        self._throttle_dns_events = {}
         self._ssl_context = ssl_context
         self._family = family
         self._local_addr = local_addr
 
+    def close(self):
+        """Close all ongoing DNS calls."""
+        for ev in self._throttle_dns_events.values():
+            ev.cancel()
+
+        super().close()
+
     @property
     def verify_ssl(self):
         """Do check for ssl certifications?"""
@@ -564,6 +646,9 @@ class TCPConnector(BaseConnector):
 
         Lazy property, creates context on demand.
         """
+        if ssl is None:  # pragma: no cover
+            raise RuntimeError('SSL is not supported.')
+
         if self._ssl_context is None:
             if not self._verify_ssl:
                 sslcontext = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
@@ -589,12 +674,12 @@ class TCPConnector(BaseConnector):
     @property
     def cached_hosts(self):
         """Read-only dict of cached DNS record."""
-        return MappingProxyType(self._cached_hosts)
+        return MappingProxyType(self._cached_hosts.addrs)
 
     def clear_dns_cache(self, host=None, port=None):
         """Remove specified host/port or clear all dns local cache."""
         if host is not None and port is not None:
-            self._cached_hosts.pop((host, port), None)
+            self._cached_hosts.remove((host, port))
         elif host is not None or port is not None:
             raise ValueError("either both host and port "
                              "or none of them are allowed")
@@ -607,18 +692,38 @@ class TCPConnector(BaseConnector):
             return [{'hostname': host, 'host': host, 'port': port,
                      'family': self._family, 'proto': 0, 'flags': 0}]
 
-        if self._use_dns_cache:
-            key = (host, port)
+        if not self._use_dns_cache:
+            return (yield from self._resolver.resolve(
+                host, port, family=self._family))
 
-            if key not in self._cached_hosts:
-                self._cached_hosts[key] = yield from \
-                    self._resolver.resolve(host, port, family=self._family)
+        key = (host, port)
 
-            return self._cached_hosts[key]
+        if (key in self._cached_hosts) and\
+                (not self._cached_hosts.expired(key)):
+            return self._cached_hosts.next_addrs(key)
+
+        if key in self._throttle_dns_events:
+            yield from self._throttle_dns_events[key].wait()
         else:
-            res = yield from self._resolver.resolve(
-                host, port, family=self._family)
-            return res
+            self._throttle_dns_events[key] = \
+                EventResultOrError(self._loop)
+            try:
+                addrs = yield from \
+                    asyncio.shield(self._resolver.resolve(host,
+                                                          port,
+                                                          family=self._family),
+                                   loop=self._loop)
+                self._cached_hosts.add(key, addrs)
+                self._throttle_dns_events[key].set()
+            except Exception as e:
+                # any DNS exception, independently of the implementation
+                # is set for the waiters to raise the same exception.
+                self._throttle_dns_events[key].set(exc=e)
+                raise
+            finally:
+                self._throttle_dns_events.pop(key)
+
+        return self._cached_hosts.next_addrs(key)
 
     @asyncio.coroutine
     def _create_connection(self, req):
@@ -633,15 +738,51 @@ class TCPConnector(BaseConnector):
 
         return proto
 
-    @asyncio.coroutine
-    def _create_direct_connection(self, req):
+    def _get_ssl_context(self, req):
+        """Logic to get the correct SSL context
+
+        0. if req.ssl is false, return None
+
+        1. if ssl_context is specified in req, use it
+        2. if _ssl_context is specified in self, use it
+        3. otherwise:
+            1. if verify_ssl is not specified in req, use self.ssl_context
+               (will generate a default context according to self.verify_ssl)
+            2. if verify_ssl is True in req, generate a default SSL context
+            3. if verify_ssl is False in req, generate a SSL context that
+               won't verify
+        """
         if req.ssl:
-            sslcontext = self.ssl_context
+            sslcontext = req.ssl_context or self._ssl_context
+            if not sslcontext:
+                if req.verify_ssl is None:
+                    sslcontext = self.ssl_context
+                elif req.verify_ssl:
+                    sslcontext = ssl.create_default_context()
+                else:
+                    sslcontext = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
+                    sslcontext.options |= ssl.OP_NO_SSLv2
+                    sslcontext.options |= ssl.OP_NO_SSLv3
+                    sslcontext.options |= _SSL_OP_NO_COMPRESSION
+                    sslcontext.set_default_verify_paths()
         else:
             sslcontext = None
+        return sslcontext
+
+    def _get_fingerprint_and_hashfunc(self, req):
+        if req.fingerprint:
+            return (req.fingerprint, req._hashfunc)
+        elif self.fingerprint:
+            return (self.fingerprint, self._hashfunc)
+        else:
+            return (None, None)
+
+    @asyncio.coroutine
+    def _create_direct_connection(self, req):
+        sslcontext = self._get_ssl_context(req)
+        fingerprint, hashfunc = self._get_fingerprint_and_hashfunc(req)
 
         hosts = yield from self._resolve_host(req.url.raw_host, req.port)
-        exc = None
 
         for hinfo in hosts:
             try:
@@ -654,7 +795,7 @@ class TCPConnector(BaseConnector):
                     server_hostname=hinfo['hostname'] if sslcontext else None,
                     local_addr=self._local_addr)
                 has_cert = transp.get_extra_info('sslcontext')
-                if has_cert and self._fingerprint:
+                if has_cert and fingerprint:
                     sock = transp.get_extra_info('socket')
                     if not hasattr(sock, 'getpeercert'):
                         # Workaround for asyncio 3.5.0
@@ -664,8 +805,8 @@ class TCPConnector(BaseConnector):
                     # gives DER-encoded cert as a sequence of bytes (or None)
                     cert = sock.getpeercert(binary_form=True)
                     assert cert
-                    got = self._hashfunc(cert).digest()
-                    expected = self._fingerprint
+                    got = hashfunc(cert).digest()
+                    expected = fingerprint
                     if got != expected:
                         transp.close()
                         if not self._cleanup_closed_disabled:
@@ -673,37 +814,45 @@ class TCPConnector(BaseConnector):
                         raise ServerFingerprintMismatch(
                             expected, got, host, port)
                 return transp, proto
-            except OSError as e:
-                exc = e
-        else:
-            raise ClientConnectorError(
-                exc.errno,
-                'Can not connect to %s:%s [%s]' %
-                (req.host, req.port, exc.strerror)) from exc
+            except certificate_errors as exc:
+                raise ClientConnectorCertificateError(
+                    req.connection_key, exc) from exc
+            except ssl_errors as exc:
+                raise ClientConnectorSSLError(req.connection_key, exc) from exc
+            except OSError as exc:
+                raise ClientConnectorError(req.connection_key, exc) from exc
 
     @asyncio.coroutine
     def _create_proxy_connection(self, req):
+        headers = {}
+        if req.proxy_headers is not None:
+            headers = req.proxy_headers
+        headers[hdrs.HOST] = req.headers[hdrs.HOST]
+
         proxy_req = ClientRequest(
             hdrs.METH_GET, req.proxy,
-            headers={hdrs.HOST: req.headers[hdrs.HOST]},
+            headers=headers,
             auth=req.proxy_auth,
-            loop=self._loop)
+            loop=self._loop,
+            verify_ssl=req.verify_ssl,
+            fingerprint=req.fingerprint,
+            ssl_context=req.ssl_context)
         try:
             # create connection to proxy server
             transport, proto = yield from self._create_direct_connection(
                 proxy_req)
         except OSError as exc:
-            raise ClientProxyConnectionError(*exc.args) from exc
+            raise ClientProxyConnectionError(proxy_req, exc) from exc
 
-        if hdrs.AUTHORIZATION in proxy_req.headers:
-            auth = proxy_req.headers[hdrs.AUTHORIZATION]
-            del proxy_req.headers[hdrs.AUTHORIZATION]
+        auth = proxy_req.headers.pop(hdrs.AUTHORIZATION, None)
+        if auth is not None:
             if not req.ssl:
                 req.headers[hdrs.PROXY_AUTHORIZATION] = auth
             else:
                 proxy_req.headers[hdrs.PROXY_AUTHORIZATION] = auth
 
         if req.ssl:
+            sslcontext = self._get_ssl_context(req)
             # For HTTPS requests over HTTP proxy
             # we must notify proxy to tunnel connection
             # so we send CONNECT command:
@@ -729,9 +878,12 @@ class TCPConnector(BaseConnector):
                 conn._transport = None
                 try:
                     if resp.status != 200:
-                        raise ClientHttpProxyError(code=resp.status,
-                                                   message=resp.reason,
-                                                   headers=resp.headers)
+                        raise ClientHttpProxyError(
+                            proxy_resp.request_info,
+                            resp.history,
+                            code=resp.status,
+                            message=resp.reason,
+                            headers=resp.headers)
                     rawsock = transport.get_extra_info('socket', default=None)
                     if rawsock is None:
                         raise RuntimeError(
@@ -742,7 +894,7 @@ class TCPConnector(BaseConnector):
                     transport.close()
 
                 transport, proto = yield from self._loop.create_connection(
-                    self._factory, ssl=self.ssl_context, sock=rawsock,
+                    self._factory, ssl=sslcontext, sock=rawsock,
                     server_hostname=req.host)
             finally:
                 proxy_resp.close()
