@@ -1,6 +1,7 @@
 # coding: utf-8
 
 import asyncio
+import hashlib
 import io
 import os.path
 import urllib.parse
@@ -263,7 +264,7 @@ def test_skip_default_useragent_header(make_request):
 
 
 def test_headers(make_request):
-    req = make_request('get', 'http://python.org/',
+    req = make_request('post', 'http://python.org/',
                        headers={'Content-Type': 'text/plain'})
 
     assert 'CONTENT-TYPE' in req.headers
@@ -272,7 +273,7 @@ def test_headers(make_request):
 
 
 def test_headers_list(make_request):
-    req = make_request('get', 'http://python.org/',
+    req = make_request('post', 'http://python.org/',
                        headers=[('Content-Type', 'text/plain')])
     assert 'CONTENT-TYPE' in req.headers
     assert req.headers['CONTENT-TYPE'] == 'text/plain'
@@ -285,15 +286,10 @@ def test_headers_default(make_request):
 
 
 def test_invalid_url(make_request):
-    with pytest.raises(ValueError):
+    with pytest.raises(aiohttp.InvalidURL):
         make_request('get', 'hiwpefhipowhefopw')
 
 
-def test_invalid_idna(make_request):
-    with pytest.raises(ValueError):
-        make_request('get', 'http://\u2061owhefopw.com')
-
-
 def test_no_path(make_request):
     req = make_request('get', 'http://python.org')
     assert '/' == req.url.path
@@ -533,16 +529,16 @@ def test_connection_header(loop, conn):
 def test_no_content_length(loop, conn):
     req = ClientRequest('get', URL('http://python.org'), loop=loop)
     resp = req.send(conn)
-    assert '0' == req.headers.get('CONTENT-LENGTH')
+    assert req.headers.get('CONTENT-LENGTH') is None
     yield from req.close()
     resp.close()
 
 
 @asyncio.coroutine
-def test_no_content_length2(loop, conn):
+def test_no_content_length_head(loop, conn):
     req = ClientRequest('head', URL('http://python.org'), loop=loop)
     resp = req.send(conn)
-    assert '0' == req.headers.get('CONTENT-LENGTH')
+    assert req.headers.get('CONTENT-LENGTH') is None
     yield from req.close()
     resp.close()
 
@@ -590,7 +586,7 @@ def test_content_type_skip_auto_header_form(loop, conn):
 
 
 def test_content_type_auto_header_content_length_no_skip(loop, conn):
-    req = ClientRequest('get', URL('http://python.org'),
+    req = ClientRequest('post', URL('http://python.org'),
                         data=io.BytesIO(b'hey'),
                         skip_auto_headers={'Content-Length'},
                         loop=loop)
@@ -630,7 +626,7 @@ def test_pass_falsy_data(loop):
         req = ClientRequest(
             'post', URL('http://python.org/'),
             data={}, loop=loop)
-        req.update_body_from_data.assert_called_once_with({}, frozenset())
+        req.update_body_from_data.assert_called_once_with({})
     yield from req.close()
 
 
@@ -649,6 +645,7 @@ def test_pass_falsy_data_file(loop, tmpdir):
     yield from req.close()
 
 
+# Elasticsearch API requires to send request body with GET-requests
 @asyncio.coroutine
 def test_get_with_data(loop):
     for meth in ClientRequest.GET_METHODS:
@@ -677,7 +674,7 @@ def test_bytes_data(loop, conn):
 
 @asyncio.coroutine
 def test_content_encoding(loop, conn):
-    req = ClientRequest('get', URL('http://python.org/'), data='foo',
+    req = ClientRequest('post', URL('http://python.org/'), data='foo',
                         compress='deflate', loop=loop)
     with mock.patch('aiohttp.client_reqrep.PayloadWriter') as m_writer:
         resp = req.send(conn)
@@ -691,7 +688,7 @@ def test_content_encoding(loop, conn):
 
 @asyncio.coroutine
 def test_content_encoding_dont_set_headers_if_no_body(loop, conn):
-    req = ClientRequest('get', URL('http://python.org/'),
+    req = ClientRequest('post', URL('http://python.org/'),
                         compress='deflate', loop=loop)
     with mock.patch('aiohttp.client_reqrep.http'):
         resp = req.send(conn)
@@ -704,7 +701,7 @@ def test_content_encoding_dont_set_headers_if_no_body(loop, conn):
 @asyncio.coroutine
 def test_content_encoding_header(loop, conn):
     req = ClientRequest(
-        'get', URL('http://python.org/'), data='foo',
+        'post', URL('http://python.org/'), data='foo',
         headers={'Content-Encoding': 'deflate'}, loop=loop)
     with mock.patch('aiohttp.client_reqrep.PayloadWriter') as m_writer:
         resp = req.send(conn)
@@ -718,7 +715,7 @@ def test_content_encoding_header(loop, conn):
 @asyncio.coroutine
 def test_compress_and_content_encoding(loop, conn):
     with pytest.raises(ValueError):
-        ClientRequest('get', URL('http://python.org/'), data='foo',
+        ClientRequest('post', URL('http://python.org/'), data='foo',
                       headers={'content-encoding': 'deflate'},
                       compress='deflate', loop=loop)
 
@@ -726,7 +723,7 @@ def test_compress_and_content_encoding(loop, conn):
 @asyncio.coroutine
 def test_chunked(loop, conn):
     req = ClientRequest(
-        'get', URL('http://python.org/'),
+        'post', URL('http://python.org/'),
         headers={'TRANSFER-ENCODING': 'gzip'}, loop=loop)
     resp = req.send(conn)
     assert 'gzip' == req.headers['TRANSFER-ENCODING']
@@ -737,7 +734,7 @@ def test_chunked(loop, conn):
 @asyncio.coroutine
 def test_chunked2(loop, conn):
     req = ClientRequest(
-        'get', URL('http://python.org/'),
+        'post', URL('http://python.org/'),
         headers={'Transfer-encoding': 'chunked'}, loop=loop)
     resp = req.send(conn)
     assert 'chunked' == req.headers['TRANSFER-ENCODING']
@@ -748,7 +745,7 @@ def test_chunked2(loop, conn):
 @asyncio.coroutine
 def test_chunked_explicit(loop, conn):
     req = ClientRequest(
-        'get', URL('http://python.org/'), chunked=True, loop=loop)
+        'post', URL('http://python.org/'), chunked=True, loop=loop)
     with mock.patch('aiohttp.client_reqrep.PayloadWriter') as m_writer:
         resp = req.send(conn)
 
@@ -762,7 +759,7 @@ def test_chunked_explicit(loop, conn):
 def test_chunked_length(loop, conn):
     with pytest.raises(ValueError):
         ClientRequest(
-            'get', URL('http://python.org/'),
+            'post', URL('http://python.org/'),
             headers={'CONTENT-LENGTH': '1000'}, chunked=True, loop=loop)
 
 
@@ -770,7 +767,7 @@ def test_chunked_length(loop, conn):
 def test_chunked_transfer_encoding(loop, conn):
     with pytest.raises(ValueError):
         ClientRequest(
-            'get', URL('http://python.org/'),
+            'post', URL('http://python.org/'),
             headers={'TRANSFER-ENCODING': 'chunked'}, chunked=True, loop=loop)
 
 
@@ -1104,7 +1101,7 @@ def test_custom_req_rep(loop):
                                        self.url,
                                        writer=self._writer,
                                        continue100=self._continue)
-            resp._post_init(self.loop)
+            resp._post_init(self.loop, mock.Mock())
             self.response = resp
             nonlocal called
             called = True
@@ -1130,3 +1127,26 @@ def test_custom_req_rep(loop):
     resp.close()
     session.close()
     conn.close()
+
+
+def test_verify_ssl_false_with_ssl_context(loop):
+    with pytest.raises(ValueError):
+        ClientRequest('get', URL('http://python.org'), verify_ssl=False,
+                      ssl_context=mock.Mock(), loop=loop)
+
+
+def test_bad_fingerprint(loop):
+    with pytest.raises(ValueError):
+        ClientRequest('get', URL('http://python.org'),
+                      fingerprint=b'invalid', loop=loop)
+
+
+def test_insecure_fingerprint(loop):
+    with pytest.warns(DeprecationWarning):
+        ClientRequest('get', URL('http://python.org'),
+                      fingerprint=hashlib.md5(b"foo").digest(),
+                      loop=loop)
+    with pytest.warns(DeprecationWarning):
+        ClientRequest('get', URL('http://python.org'),
+                      fingerprint=hashlib.sha1(b"foo").digest(),
+                      loop=loop)
