@@ -1,12 +1,13 @@
 import asyncio
 import ipaddress
 import socket
-from unittest.mock import patch
+from unittest.mock import Mock, patch
 
 import pytest
 
 from aiohttp.resolver import AsyncResolver, DefaultResolver, ThreadedResolver
 
+
 try:
     import aiodns
     gethostbyname = hasattr(aiodns.DNSResolver, 'gethostbyname')
@@ -102,7 +103,7 @@ def test_async_resolver_negative_lookup(loop):
     with patch('aiodns.DNSResolver') as mock:
         mock().gethostbyname.side_effect = aiodns.error.DNSError()
         resolver = AsyncResolver(loop=loop)
-        with pytest.raises(aiodns.error.DNSError):
+        with pytest.raises(OSError):
             yield from resolver.resolve('doesnotexist.bla')
 
 
@@ -113,12 +114,34 @@ def test_async_resolver_query_negative_lookup(loop):
         del mock().gethostbyname
         mock().query.side_effect = aiodns.error.DNSError()
         resolver = AsyncResolver(loop=loop)
-        with pytest.raises(aiodns.error.DNSError):
+        with pytest.raises(OSError):
+            yield from resolver.resolve('doesnotexist.bla')
+
+
+@pytest.mark.skipif(aiodns is None, reason="aiodns required")
+@asyncio.coroutine
+def test_async_resolver_no_hosts_in_query(loop):
+    with patch('aiodns.DNSResolver') as mock:
+        del mock().gethostbyname
+        mock().query.return_value = fake_query_result([])
+        resolver = AsyncResolver(loop=loop)
+        with pytest.raises(OSError):
+            yield from resolver.resolve('doesnotexist.bla')
+
+
+@pytest.mark.skipif(not gethostbyname, reason="aiodns 1.1 required")
+@asyncio.coroutine
+def test_async_resolver_no_hosts_in_gethostbyname(loop):
+    with patch('aiodns.DNSResolver') as mock:
+        mock().gethostbyname.return_value = fake_result([])
+        resolver = AsyncResolver(loop=loop)
+        with pytest.raises(OSError):
             yield from resolver.resolve('doesnotexist.bla')
 
 
 @asyncio.coroutine
-def test_threaded_resolver_positive_lookup(loop):
+def test_threaded_resolver_positive_lookup():
+    loop = Mock()
     loop.getaddrinfo = fake_addrinfo(["127.0.0.1"])
     resolver = ThreadedResolver(loop=loop)
     real = yield from resolver.resolve('www.python.org')
@@ -126,7 +149,8 @@ def test_threaded_resolver_positive_lookup(loop):
 
 
 @asyncio.coroutine
-def test_threaded_resolver_multiple_replies(loop):
+def test_threaded_resolver_multiple_replies():
+    loop = Mock()
     ips = ['127.0.0.1', '127.0.0.2', '127.0.0.3', '127.0.0.4']
     loop.getaddrinfo = fake_addrinfo(ips)
     resolver = ThreadedResolver(loop=loop)
@@ -136,7 +160,8 @@ def test_threaded_resolver_multiple_replies(loop):
 
 
 @asyncio.coroutine
-def test_threaded_negative_lookup(loop):
+def test_threaded_negative_lookup():
+    loop = Mock()
     ips = []
     loop.getaddrinfo = fake_addrinfo(ips)
     resolver = ThreadedResolver(loop=loop)
@@ -203,7 +228,8 @@ def test_async_resolver_aiodns_not_present(loop, monkeypatch):
 
 
 def test_default_resolver():
-    if gethostbyname:
-        assert DefaultResolver is AsyncResolver
-    else:
-        assert DefaultResolver is ThreadedResolver
+    # if gethostbyname:
+    #     assert DefaultResolver is AsyncResolver
+    # else:
+    #     assert DefaultResolver is ThreadedResolver
+    assert DefaultResolver is ThreadedResolver
