@@ -12,6 +12,7 @@ from yarl import URL
 import aiohttp
 from aiohttp import FormData, HttpVersion10, HttpVersion11, multipart, web
 
+
 try:
     import ssl
 except:
@@ -525,7 +526,7 @@ def test_100_continue_custom_response(loop, test_client):
             if auth_err:
                 return web.HTTPForbidden()
 
-            request.transport.write(b"HTTP/1.1 100 Continue\r\n\r\n")
+            request.writer.write(b"HTTP/1.1 100 Continue\r\n\r\n")
 
     form = FormData()
     form.add_field('name', b'123',
@@ -693,14 +694,16 @@ def test_empty_content_for_query_without_body(loop, test_client):
 
     @asyncio.coroutine
     def handler(request):
+        assert not request.body_exists
+        assert not request.can_read_body
         assert not request.has_body
         return web.Response()
 
     app = web.Application()
-    app.router.add_get('/', handler)
+    app.router.add_post('/', handler)
     client = yield from test_client(app)
 
-    resp = yield from client.get('/')
+    resp = yield from client.post('/')
     assert 200 == resp.status
 
 
@@ -709,6 +712,8 @@ def test_empty_content_for_query_with_body(loop, test_client):
 
     @asyncio.coroutine
     def handler(request):
+        assert request.body_exists
+        assert request.can_read_body
         assert request.has_body
         body = yield from request.read()
         return web.Response(body=body)
@@ -762,12 +767,12 @@ def test_large_header_allowed(loop, test_client, test_server):
         return web.Response()
 
     app = web.Application()
-    app.router.add_get('/', handler)
+    app.router.add_post('/', handler)
     server = yield from test_server(app, max_field_size=81920)
     client = yield from test_client(server)
 
     headers = {'Long-Header': 'ab' * 8129}
-    resp = yield from client.get('/', headers=headers)
+    resp = yield from client.post('/', headers=headers)
     assert 200 == resp.status
 
 
@@ -992,7 +997,7 @@ def test_bad_request_payload(loop, test_client):
 
     @asyncio.coroutine
     def handler(request):
-        assert request.method == 'GET'
+        assert request.method == 'POST'
 
         with pytest.raises(aiohttp.web.RequestPayloadError):
             yield from request.content.read()
@@ -1000,10 +1005,10 @@ def test_bad_request_payload(loop, test_client):
         return web.Response()
 
     app = web.Application()
-    app.router.add_get('/', handler)
+    app.router.add_post('/', handler)
     client = yield from test_client(app)
 
-    resp = yield from client.get(
+    resp = yield from client.post(
         '/', data=b'test', headers={'content-encoding': 'gzip'})
     assert 200 == resp.status
 
@@ -1565,3 +1570,21 @@ def test_response_with_bodypart(loop, test_client):
         resp.headers['content-disposition'])
     assert disp == ('attachment',
                     {'name': 'file', 'filename': 'file', 'filename*': 'file'})
+
+
+@asyncio.coroutine
+def test_request_clone(loop, test_client):
+
+    @asyncio.coroutine
+    def handler(request):
+        r2 = request.clone(method='POST')
+        assert r2.method == 'POST'
+        assert r2.match_info is request.match_info
+        return web.Response()
+
+    app = web.Application()
+    app.router.add_get('/', handler)
+    client = yield from test_client(app)
+
+    resp = yield from client.get('/')
+    assert 200 == resp.status
