@@ -7,25 +7,21 @@ from aiohttp import web
 
 @asyncio.coroutine
 def test_middleware_modifies_response(loop, test_client):
-
     @asyncio.coroutine
     def handler(request):
         return web.Response(body=b'OK')
 
     @asyncio.coroutine
-    def middleware_factory(app, handler):
-
-        @asyncio.coroutine
-        def middleware(request):
-            resp = yield from handler(request)
-            assert 200 == resp.status
-            resp.set_status(201)
-            resp.text = resp.text + '[MIDDLEWARE]'
-            return resp
-        return middleware
+    @web.middleware
+    def middleware(request, handler):
+        resp = yield from handler(request)
+        assert 200 == resp.status
+        resp.set_status(201)
+        resp.text = resp.text + '[MIDDLEWARE]'
+        return resp
 
     app = web.Application()
-    app.middlewares.append(middleware_factory)
+    app.middlewares.append(middleware)
     app.router.add_route('GET', '/', handler)
     client = yield from test_client(app)
     resp = yield from client.get('/')
@@ -36,25 +32,20 @@ def test_middleware_modifies_response(loop, test_client):
 
 @asyncio.coroutine
 def test_middleware_handles_exception(loop, test_client):
-
     @asyncio.coroutine
     def handler(request):
         raise RuntimeError('Error text')
 
     @asyncio.coroutine
-    def middleware_factory(app, handler):
-
-        @asyncio.coroutine
-        def middleware(request):
-            with pytest.raises(RuntimeError) as ctx:
-                yield from handler(request)
-            return web.Response(status=501,
-                                text=str(ctx.value) + '[MIDDLEWARE]')
-
-        return middleware
+    @web.middleware
+    def middleware(request, handler):
+        with pytest.raises(RuntimeError) as ctx:
+            yield from handler(request)
+        return web.Response(status=501,
+                            text=str(ctx.value) + '[MIDDLEWARE]')
 
     app = web.Application()
-    app.middlewares.append(middleware_factory)
+    app.middlewares.append(middleware)
     app.router.add_route('GET', '/', handler)
     client = yield from test_client(app)
     resp = yield from client.get('/')
@@ -65,27 +56,22 @@ def test_middleware_handles_exception(loop, test_client):
 
 @asyncio.coroutine
 def test_middleware_chain(loop, test_client):
-
     @asyncio.coroutine
     def handler(request):
         return web.Response(text='OK')
 
-    def make_factory(num):
-
+    def make_middleware(num):
         @asyncio.coroutine
-        def factory(app, handler):
-
-            def middleware(request):
-                resp = yield from handler(request)
-                resp.text = resp.text + '[{}]'.format(num)
-                return resp
-
-            return middleware
-        return factory
+        @web.middleware
+        def middleware(request, handler):
+            resp = yield from handler(request)
+            resp.text = resp.text + '[{}]'.format(num)
+            return resp
+        return middleware
 
     app = web.Application()
-    app.middlewares.append(make_factory(1))
-    app.middlewares.append(make_factory(2))
+    app.middlewares.append(make_middleware(1))
+    app.middlewares.append(make_middleware(2))
     app.router.add_route('GET', '/', handler)
     client = yield from test_client(app)
     resp = yield from client.get('/')
@@ -118,7 +104,11 @@ class TestNormalizePathMiddleware:
         ('/resource1', 200),
         ('/resource1/', 404),
         ('/resource2', 200),
-        ('/resource2/', 200)
+        ('/resource2/', 200),
+        ('/resource1?p1=1&p2=2', 200),
+        ('/resource1/?p1=1&p2=2', 404),
+        ('/resource2?p1=1&p2=2', 200),
+        ('/resource2/?p1=1&p2=2', 200)
     ])
     def test_add_trailing_when_necessary(
             self, path, status, cli):
@@ -134,7 +124,11 @@ class TestNormalizePathMiddleware:
         ('/resource1', 200),
         ('/resource1/', 404),
         ('/resource2', 404),
-        ('/resource2/', 200)
+        ('/resource2/', 200),
+        ('/resource1?p1=1&p2=2', 200),
+        ('/resource1/?p1=1&p2=2', 404),
+        ('/resource2?p1=1&p2=2', 404),
+        ('/resource2/?p1=1&p2=2', 200)
     ])
     def test_no_trailing_slash_when_disabled(
             self, path, status, cli):
@@ -153,7 +147,13 @@ class TestNormalizePathMiddleware:
         ('//resource1//a//b/', 404),
         ('///resource1//a//b', 200),
         ('/////resource1/a///b', 200),
-        ('/////resource1/a//b/', 404)
+        ('/////resource1/a//b/', 404),
+        ('/resource1/a/b?p=1', 200),
+        ('//resource1//a//b?p=1', 200),
+        ('//resource1//a//b/?p=1', 404),
+        ('///resource1//a//b?p=1', 200),
+        ('/////resource1/a///b?p=1', 200),
+        ('/////resource1/a//b/?p=1', 404),
     ])
     def test_merge_slash(self, path, status, cli):
         extra_middlewares = [
@@ -179,7 +179,22 @@ class TestNormalizePathMiddleware:
         ('///resource2//a//b', 200),
         ('///resource2//a//b/', 200),
         ('/////resource2/a///b', 200),
-        ('/////resource2/a///b/', 200)
+        ('/////resource2/a///b/', 200),
+        ('/resource1/a/b?p=1', 200),
+        ('/resource1/a/b/?p=1', 404),
+        ('//resource2//a//b?p=1', 200),
+        ('//resource2//a//b/?p=1', 200),
+        ('///resource1//a//b?p=1', 200),
+        ('///resource1//a//b/?p=1', 404),
+        ('/////resource1/a///b?p=1', 200),
+        ('/////resource1/a///b/?p=1', 404),
+        ('/resource2/a/b?p=1', 200),
+        ('//resource2//a//b?p=1', 200),
+        ('//resource2//a//b/?p=1', 200),
+        ('///resource2//a//b?p=1', 200),
+        ('///resource2//a//b/?p=1', 200),
+        ('/////resource2/a///b?p=1', 200),
+        ('/////resource2/a///b/?p=1', 200)
     ])
     def test_append_and_merge_slash(self, path, status, cli):
         extra_middlewares = [
@@ -188,3 +203,90 @@ class TestNormalizePathMiddleware:
         client = yield from cli(extra_middlewares)
         resp = yield from client.get(path)
         assert resp.status == status
+
+
+@asyncio.coroutine
+def test_old_style_middleware(loop, test_client):
+    @asyncio.coroutine
+    def handler(request):
+        return web.Response(body=b'OK')
+
+    @asyncio.coroutine
+    def middleware_factory(app, handler):
+
+        @asyncio.coroutine
+        def middleware(request):
+            resp = yield from handler(request)
+            assert 200 == resp.status
+            resp.set_status(201)
+            resp.text = resp.text + '[old style middleware]'
+            return resp
+        return middleware
+
+    with pytest.warns(DeprecationWarning) as warning_checker:
+        app = web.Application()
+        app.middlewares.append(middleware_factory)
+        app.router.add_route('GET', '/', handler)
+        client = yield from test_client(app)
+        resp = yield from client.get('/')
+        assert 201 == resp.status
+        txt = yield from resp.text()
+        assert 'OK[old style middleware]' == txt
+
+    assert len(warning_checker) == 1
+    msg = str(warning_checker.list[0].message)
+    assert 'old-style middleware "middleware_factory" deprecated' in msg
+
+
+@asyncio.coroutine
+def test_mixed_middleware(loop, test_client):
+    @asyncio.coroutine
+    def handler(request):
+        return web.Response(body=b'OK')
+
+    @asyncio.coroutine
+    def m_old1(app, handler):
+        @asyncio.coroutine
+        def middleware(request):
+            resp = yield from handler(request)
+            resp.text += '[old style 1]'
+            return resp
+        return middleware
+
+    @asyncio.coroutine
+    @web.middleware
+    def m_new1(request, handler):
+        resp = yield from handler(request)
+        resp.text += '[new style 1]'
+        return resp
+
+    @asyncio.coroutine
+    def m_old2(app, handler):
+        @asyncio.coroutine
+        def middleware(request):
+            resp = yield from handler(request)
+            resp.text += '[old style 2]'
+            return resp
+        return middleware
+
+    @asyncio.coroutine
+    @web.middleware
+    def m_new2(request, handler):
+        resp = yield from handler(request)
+        resp.text += '[new style 2]'
+        return resp
+
+    middlewares = m_old1, m_new1, m_old2, m_new2
+
+    with pytest.warns(DeprecationWarning) as w:
+        app = web.Application(middlewares=middlewares)
+        app.router.add_route('GET', '/', handler)
+        client = yield from test_client(app)
+        resp = yield from client.get('/')
+        assert 200 == resp.status
+        txt = yield from resp.text()
+        assert 'OK[new style 2][old style 2][new style 1][old style 1]' == txt
+
+    assert len(w) == 2
+    assert 'old-style middleware "m_old2" deprecated' in str(w.list[0].message)
+    assert 'old-style middleware "m_old1" deprecated' in str(w.list[1].message)
